/*
 * FILE --- ga.cc
 * 
 * PURPOSE --- Header file for a utility that implements a Genetic 
 * Algorithm (GA) that is a stochastic direct search or optimization method.
 * It manipulates a string of numbers (a chromosome) in a manner similar to 
 * the laws of evolution and biology.
 *
 */

// Include Files
#include "ga.hh"
#include "random.hh"

// Define this macro if you want to use this utility without linking
// to the RCS Library
#ifdef NO_RCS
#include <stdio.h>            /* printf function */
#else
#include "rcs_prnt.hh"          /* rcs_print_ functions */
#endif

/*************************************************************/

FITNESS_PARAM::FITNESS_PARAM()
{
  // Allocate here memory if needed

  // Initialize the parameters
  example = 0.0;

}

FITNESS_PARAM::~FITNESS_PARAM()
{
  // Free here memory if needed

}


// Chromosome is a string of genes that will be contained in each 
// individual
CHROMOSOME::CHROMOSOME(int _chrom_length)
{
  // No error at start
  chrom_err = GA_NO_ERROR;

  // Initialize parameters
  chrom_length = _chrom_length;

  // Allocate memory
  if ((gene = new int[chrom_length]) == NULL)
    {
      chrom_err = GA_MEM_ERROR;
#ifdef NO_RCS
      printf("CHROMOSOME: ERROR --- Not Enought Memory!!!\n");
      printf("The Genetic Algorithm will not function correctly ...\n");
#else
      rcs_print("CHROMOSOME: ERROR --- Not Enought Memory!!!\n");
      rcs_print("The Genetic Algorithm will not function correctly ...\n");
#endif
    }
}

CHROMOSOME::~CHROMOSOME()
{
  // Free the allocated memory
  delete [] gene;
}

void CHROMOSOME::set_chrom(int *_genes)
{
  // Set the choromosome to the specified values
  // the array _genes should be of length chrom_length
  for(int ii = 0; ii < chrom_length; ii++)
    {
      gene[ii] = _genes[ii];
    }
}

void CHROMOSOME::create_chrom(int *_gene_low, int *_gene_high)
{
  // Create a new cromosome with genes b/n the specified boundaries
  // The arrays _gene_low and _gene_high should be of length chrom_length
  for(int ii = 0; ii < chrom_length; ii++)
    {
      gene[ii] = Rand_Int(_gene_low[ii], _gene_high[ii]);
    }
} 

void CHROMOSOME::mutate(int _gene, int _low, int _high)
{
  // Mutate the specified gene within the specified boundaries
  gene[_gene] = Rand_Int(_low, _high);
}

// Traits are decimal numbers that are decoded from the sections
// of each chromosome
TRAITS::TRAITS(int _num_traits)
{
  // No error at start
  trait_err = GA_NO_ERROR;

  // Initialize parameters
  num_traits = _num_traits;

  // Allocate memory 
  if ((trait_val = new double[num_traits]) == NULL)
    {
      trait_err = GA_MEM_ERROR;
#ifdef NO_RCS
      printf("TRAITS: ERROR --- Not Enought Memory!!!\n");
      printf("The Genetic Algorithm will not function correctly ...\n");
#else
      rcs_print("TRAITS: ERROR --- Not Enought Memory!!!\n");
      rcs_print("The Genetic Algorithm will not function correctly ...\n");
#endif
    }
}

TRAITS::~TRAITS()
{
  // Free the allocated memory
  delete [] trait_val;
}

INDIVIDUAL::INDIVIDUAL(int _chrom_length, int _num_traits) 
  : CHROMOSOME(_chrom_length), TRAITS(_num_traits)
{
  // No error at start
  indv_err = GA_NO_ERROR;

  // Initialize parameters
  fitness = 0.0;
  parents[0] = 0;
  parents[1] = 0;

  // Check if error has occured in the base classes
  if ((chrom_err != GA_NO_ERROR) || (trait_err != GA_NO_ERROR))
    {
      indv_err = GA_MEM_ERROR;
#ifdef NO_RCS
      printf("INDIVIDUAL: ERROR --- Not Enought Memory!!!\n");
      printf("The Genetic Algorithm will not function correctly ...\n");
#else
      rcs_print("INDIVIDUAL: ERROR --- Not Enought Memory!!!\n");
      rcs_print("The Genetic Algorithm will not function correctly ...\n");
#endif
    }
}

INDIVIDUAL::~INDIVIDUAL()
{

}

// A survivor is an individual who is selected to reproduce into the next 
// generation. In other words, its genetic material (chromosome) will 
// be passed to the the next generations.
SURVIVOR::SURVIVOR(int _chrom_length)
  : CHROMOSOME(_chrom_length)
{
  // No error at start
  surv_err = GA_NO_ERROR;

  // Initialize parameters
  surv_num = 0;
  
  // Check if error has occured in the base class
  if (chrom_err != GA_NO_ERROR)
    {
      surv_err = GA_MEM_ERROR;
#ifdef NO_RCS
      printf("SURVIVOR: ERROR --- Not Enought Memory!!!\n");
      printf("The Genetic Algorithm will not function correctly ...\n");
#else
      rcs_print("SURVIVOR: ERROR --- Not Enought Memory!!!\n");
      rcs_print("The Genetic Algorithm will not function correctly ...\n");
#endif
    }
}

SURVIVOR::~SURVIVOR()
{

}


POPULATION::POPULATION(int _pop_size, int _chrom_length, int _num_traits)
{
  // No error at start
  pop_err = GA_NO_ERROR;

  // Initialize parameters
  pop_size = _pop_size;
  chrom_length = _chrom_length;
  num_traits = _num_traits;

  sumfitness = 0.0;
  bestmember = 0;
  worstmember = 0;

  generation = 0;
  crossovers = 0;
  mutations = 0;

  // Allow only even population
  if (pop_size % 2)
    {
      pop_size++;
    }
    
  // Allocate memory 
  if ((members = new Individual*[pop_size]) == NULL)
    pop_err = GA_MEM_ERROR;

  if ((survivors = new Survivor*[pop_size]) == NULL)
    pop_err = GA_MEM_ERROR;
  
  for(int ii = 0; ii < pop_size; ii++)
    {
      if ((members[ii] = new Individual(chrom_length, num_traits)) == NULL)
	pop_err = GA_MEM_ERROR;

      if (members[ii]->indv_err != GA_NO_ERROR)
	pop_err = GA_MEM_ERROR;

      if ((survivors[ii] = new Survivor(chrom_length)) == NULL)
	pop_err = GA_MEM_ERROR;

      if (survivors[ii]->surv_err != GA_NO_ERROR)
	pop_err = GA_MEM_ERROR;
    }
  
  if ((trait_start = new int[_num_traits+1]) == NULL)
    pop_err = GA_MEM_ERROR;
  
  if ((trait_offset = new double[_num_traits]) == NULL)
    pop_err = GA_MEM_ERROR;

  if ((trait_low = new double[_num_traits]) == NULL)
    pop_err = GA_MEM_ERROR;

  if ((trait_high = new double[_num_traits]) == NULL)
    pop_err = GA_MEM_ERROR;

  if ((decimal = new int[_num_traits]) == NULL)
    pop_err = GA_MEM_ERROR;

  if ((gene_low = new int[_chrom_length]) == NULL)
    pop_err = GA_MEM_ERROR;

  if ((gene_high = new int[_chrom_length]) == NULL)
    pop_err = GA_MEM_ERROR;

  // Check if error has occured   
  if (pop_err != GA_NO_ERROR)
    {
#ifdef NO_RCS
      printf("POPULATION: ERROR --- Not Enought Memory!!!\n");
      printf("The Genetic Algorithm will not function correctly ...\n");
#else
      rcs_print("POPULATION: ERROR --- Not Enought Memory!!!\n");
      rcs_print("The Genetic Algorithm will not function correctly ...\n");
#endif
    }
}

POPULATION::~POPULATION()
{
  // Free the allocated memory
  for(int ii = 0; ii < pop_size; ii++)
    {
      delete members[ii];
      delete survivors[ii];
    }
  delete [] members;
  delete [] survivors;

  delete [] trait_start;
  delete [] trait_offset;
  delete [] trait_low;
  delete [] trait_high;
  delete [] decimal;
  delete [] gene_low;
  delete [] gene_high;
}

void POPULATION::set_trait_start(int *_trait_start)
{
  // Set the boundaries of the traits within the chromosome
  for(int ii = 0; ii < num_traits; ii++)
    {
      trait_start[ii] = _trait_start[ii];
    }
  trait_start[num_traits] = chrom_length;
}

void POPULATION::set_trait_offset(double *_trait_offset)
{
  // Set the trait offsets to be used in decoding the chromosomes
  for(int ii = 0; ii < num_traits; ii++)
    {
      trait_offset[ii] = _trait_offset[ii];
    }
}

void POPULATION::set_trait_low(double *_trait_low)
{
  // Set the valid lower ranges of the traits 
  for(int ii = 0; ii < num_traits; ii++)
    {
      trait_low[ii] = _trait_low[ii];
    }
}

void POPULATION::set_trait_high(double *_trait_high)
{
  // Set the valid upper ranges of the traits 
  for(int ii = 0; ii < num_traits; ii++)
    {
      trait_high[ii] = _trait_high[ii];
    }
}

void POPULATION::set_decimal(int *_decimal)
{
  // Set the place of the decimal point to be used in decoding the chromosomes
  for(int ii = 0; ii < num_traits; ii++)
    {
      decimal[ii] = _decimal[ii];
    }
}

void POPULATION::set_gene_low(int *_gene_low)
{
  // Set the valid lower ranges of the genes for the chromosoms
  for(int ii = 0; ii < chrom_length; ii++)
    {
      gene_low[ii] = _gene_low[ii];
    }
}

void POPULATION::set_gene_high(int *_gene_high)
{
  // Set the valid upper ranges of the genes for the chromosoms
  for(int ii = 0; ii < chrom_length; ii++)
    {
      gene_high[ii] = _gene_high[ii];
    }
}

void POPULATION::set_cross_type(int _cross_type)
{
  // Set the crossover type 1 or 0
  // 0 = swap each gene with 50% probability
  // 1 = swap all the genes before a random selected site
  cross_type = _cross_type;  

  // Make sure that the type is correct
  if (cross_type < 0)
    {
      // swap each gene with 50% probability
      cross_type = 0; 
    }
  else if (cross_type > 1)
    {
      // swap all the genes before a random selected site
      cross_type = 1; 
    }
}

void POPULATION::set_cross_prob(double _cross_prob)
{
  // Set the crossover and mutation probabilities
  cross_prob = _cross_prob;  

  // Make sure that the probability is correct
  if (cross_prob < 0.0)
    {
      cross_prob = 0.0; 
    }
  else if (cross_prob > 1.0)
    {
      cross_prob = 1.0; 
    }
}

void POPULATION::set_mutat_prob(double _mutat_prob)
{
  mutat_prob = _mutat_prob;  

  // Make sure that the probability is correct
  if (mutat_prob < 0.0)
    {
      mutat_prob = 0.0; 
    }
  else if (mutat_prob > 1.0)
    {
      mutat_prob = 1.0; 
    }
}

void POPULATION::set_elit_type(int _elit_type)
{
  // Use elitism or not 1=on 0=off
  use_elitism = _elit_type;

  // Make sure that the type is correct
  if (use_elitism < 0)
    {
      use_elitism = 0; // Elitism off
    }
  else if (use_elitism > 1)
    {
      use_elitism = 1; // Elitism on
    }
}

void POPULATION::set_decode_type(int _decode_type)
{
  // Use sign in decoding traits or not 1=signed 0=unsigned
  // If signed traits are used the first gene of each chromosome
  // determines the sign of the trait.
  // If this gene is 0-4, the trait is negative.
  // If this gene is 5-9, the trait is positive.
  use_sign = _decode_type;

  // Make sure that the type is correct
  if (use_sign < 0)
    {
      use_sign = 0; // Traits are unsigned
    }
  else if (use_sign > 1)
    {
      use_sign = 1; // Traits are signed
    }
}

void POPULATION::decode_chrom(Individual *individual)
{
  // Decodes the chromosome of the individual into traits.

  int ii, jj; // Index variables for the loops

  int place, index;
  double temp;

  // If (use_sign==1) then the first gene of each trait is the 
  // sign of the trait. The trait is + if this gene is 5-9 and - if 0-4
  for(ii = 0; ii < num_traits; ii++)
    { 
      // start with zero value and the first gene of the trait
      temp = 0.0;
      index = 1;
      for(jj = trait_start[ii]+use_sign; jj < trait_start[ii+1]; jj++)
        {
	  // Find the place before or after the decimal point
          place = decimal[ii] - index;
	  
	  // Add the contribution of the gene
          temp += individual->gene[jj]*Power(10,place);

	  // Next gene
          index++;
        }

      // If signed traits find the sign
      if (use_sign && (individual->gene[trait_start[ii]] < 5)) 
	temp = - temp;

      // Add the offset
      individual->trait_val[ii] = temp + trait_offset[ii];
    }
}

int POPULATION::check_chrom(Individual *individual)
{
  // Checks the chromosome of the individual. Returns 0 if chrom 
  // has all legal genes, otherwise returns the number of the last 
  // bad trait number it finds.

  int flag; // Variable to keep the numbe of the bad trait

  flag = 0; // No bad trait at start

  // Make sure that traits are current
  decode_chrom(individual);
  
  // Check whether within the valid ranges
  for(int ii = 0; ii < num_traits; ii++)
    {
    if ((individual->trait_val[ii] > trait_high[ii]) || 
        (individual->trait_val[ii] < trait_low[ii]))
      flag = ii + 1;
    }
  
  return (flag);
}

void POPULATION::fix_chrom(Individual *individual, int _flag)
{
  // Fix the chromosome (make it valid)
  // _flag specifies the trait to be fixed

  // Make sure that traits are current
  decode_chrom(individual);

  // Randomly create a new, valid, trait.
  individual->trait_val[_flag-1] = 
    Rand_Double(trait_low[_flag-1], trait_high[_flag-1]);
  
  // Encode the traits back into the chromosome
  encode_traits(individual);
}


void POPULATION::encode_traits(Individual *individual)
{
  // Encodes the traits of the individual into chromosome.
  
  int ii, jj; // Index variables for the loops

  int gene;
  double temp;
  double pow1;
  double pow2;
  

  for(ii = 0; ii < num_traits; ii++)
    {
      temp = individual->trait_val[ii] - trait_offset[ii];
      
      // If (use_sign == 1) account for +/- sign
      if (use_sign)
        {
          if (temp < 0.0) 
	    {
	      individual->gene[trait_start[ii]] = 0;
	    }
          else
	    { 
	      individual->gene[trait_start[ii]] = 9;
	    }
        }
      temp = fabs(temp);
      
      pow1 = Power(10,1-decimal[ii]);
      pow2 = Power(10,decimal[ii]-1);
      
      // Encode the rest of the trait.
      for(jj = trait_start[ii]+use_sign; jj < trait_start[ii+1]; jj++)
        {
          gene = (int) (temp * pow1 + ZERO);
          individual->gene[jj] = gene;
          temp = (temp - (double)gene * pow2) * 10.0;
        }
    }
}


void POPULATION::find_fitness(Fitness_param *_fitness_param)
{
  // Find the fitness of all the individuals in the population.

  int ii; // Index variables for the loops
  double fitness, bestfitness, worstfitness;


  // Get the fitness of the first individual
  decode_chrom(members[0]);
  //  members[0]->get_fitness(_fitness_param);
  get_fitness(members[0], _fitness_param);
  sumfitness  =  members[0]->fitness;
  
  // Currently this is both the best and worst individual
  bestfitness = members[0]->fitness;
  worstfitness  = members[0]->fitness;
  bestmember = 0;
  worstmember = 0;

  
  // Find the fitness of the rest
  for(ii = 1; ii < pop_size; ii++)
    {
      decode_chrom(members[ii]);
      //      members[ii]->get_fitness(_fitness_param);
      get_fitness(members[ii], _fitness_param);
      fitness  =  members[ii]->fitness;
      sumfitness += fitness;
      
      // Find the best member
      if (fitness > bestfitness)
        {
          bestfitness = fitness;
          bestmember = ii;
        }
      
      // Find the worst member
      if (fitness < worstfitness)
        {
          worstfitness = fitness;
          worstmember = ii;
        }
    }
}

void POPULATION::first_generation()
{
  // Create a random first generation
  
  for(int ii = 0; ii < pop_size; ii++)
    {
      // Create a valid chromosome
      do {
	members[ii]->create_chrom(gene_low, gene_high);
      } while (check_chrom(members[ii]));
      
      // Parents are unknown (actually zeroth individual)
      members[ii]->parents[0] = 0;
      members[ii]->parents[1] = 0;
    }
  
  generation = 1;
  crossovers = 0;
  mutations = 0;
}

void POPULATION::next_generation(void)
{
  // Given a generation, create the next generation
  // This is done by selecting survivors, creating the offspring of those
  // survivors.

  // No crossovers or mutations at this moment
  crossovers = 0;
  mutations = 0;

  // Select the lucky survivors
  select_survivors();

  // Mate the survivors and make offspring
  make_offspring();

  // We are at the next generation now
  generation++;
}  

void POPULATION::select_survivors(void)
{
  // Select the survivors from the given generation
  // The survivors are the individuals which will be reproduced
  // to form the next generation (mate and make offspring)
  
  // This function uses the 'Roulette-Wheel' selection method 
  // as described by David E. Goldberg.

  int member_count;
  int count;
  double pointer;
  double sum;


  // DETERMINE NUMBER OF PARENTS TO CREATE
  if (use_elitism)
    {
      count = pop_size - 1;
    }
  else
    {
      count = ((pop_size+1)/2)*2;
    }

  // Determine which member is selected by the roulette wheel
  for(int ii = 0; ii < count; ii++)
    {  
      pointer = sumfitness*Rand_Num();

      member_count = - 1;
      sum = 0.0;
      
      while (sum < pointer)
        {
          member_count++;
          
          // Check for error
          if (member_count > pop_size)
            {
#ifdef NO_RCS
	      printf("select_survivors: ERROR Negative Fitness!\n");
#else
 	      rcs_print("select_survivors: ERROR Negative Fitness!\n");
#endif
	      pop_err = GA_MEM_NEGATIVE_FITNESS;
	      member_count = 0;
            }
          
          sum += members[member_count]->fitness;
        }
      
      // The selected member will survive and become parent
      survivors[ii]->set_chrom(members[member_count]->gene);
      survivors[ii]->surv_num = member_count;
    }
  
  // If elitism is on then add also the best individual
  if (use_elitism)
    {
      survivors[pop_size-1]->set_chrom(members[bestmember]->gene);
      survivors[pop_size-1]->surv_num = bestmember;
    }
}

void POPULATION::make_offspring(void)
{
  // Make the offspring from the survivors - create a new generation
  // of inidividuals. This is done by mating survivor[n] with 
  // survivor[n+1], n=0,2,4, etc.

  int ii; // Index variables for the loops

  // Mate the neighboring survivors
  for(ii = 0; ii < pop_size-use_elitism; ii += 2)
    {      
      members[ii]->set_chrom(survivors[ii]->gene);
      members[ii]->parents[0] = survivors[ii]->surv_num;
      members[ii]->parents[1] = survivors[ii+1]->surv_num;

      members[ii+1]->set_chrom(survivors[ii+1]->gene);
      members[ii+1]->parents[0] = survivors[ii]->surv_num;
      members[ii+1]->parents[1] = survivors[ii+1]->surv_num;

      if (Flip(cross_prob))
	{
	  crossover(members[ii], members[ii+1]);
	  crossovers++;
	}
      
      mutate(members[ii]);
      mutate(members[ii+1]);
    }
 
  // If elitism is on the best individual "clones" himself
  // In other words survives without any change in its cromosome
  if (use_elitism)
    {
      members[pop_size-1]->set_chrom(survivors[pop_size-1]->gene);
      members[pop_size-1]->parents[0] = survivors[pop_size-1]->surv_num;
      members[pop_size-1]->parents[1] = survivors[pop_size-1]->surv_num;
    }
}

void POPULATION::crossover(Individual *_individual1, Individual *_individual2)
{
  // Mate (make crossover) _individual1 with _individual2
  // If this creates an "illegal" chromosome, then the 'nearest' 'legal' 
  // chromosome is created in its place

  int ii; // Index variables for the loops
  int temp;
  int flag;
  int site;


  // cross_type == 0 => do crossover on a gene by gene basis
  // cross_type == 1 => select a site and flip all the genes
  // before that site
  if (cross_type == 0)
    {
      for(ii = 0; ii < chrom_length; ii++)
        if (Flip(0.5))
          {
	    temp = _individual1->gene[ii];
            _individual1->gene[ii] = _individual2->gene[ii];
            _individual2->gene[ii] = temp;
          }
    }
  else
    {
      site = Rand_Int(1, chrom_length);
      
      for(ii = 0; ii < site; ii++)
        {
	  temp = _individual1->gene[ii];
	  _individual1->gene[ii] = _individual2->gene[ii];
	  _individual2->gene[ii] = temp;
        }
    }


  // Make sure that the new individuals (children are valid)
  do 
    {
      flag = check_chrom(_individual1);
      if (flag) fix_chrom(_individual1, flag);
    } while (flag);

  do 
    {
      flag = check_chrom(_individual2);
      if (flag) fix_chrom(_individual2, flag);
    } while (flag);
}

void POPULATION::mutate(Individual *individual)
{
  // Mutate the chromosome of the given individual

  int ii; // Index variables for the loops
  int flag;

  // Mutate the cromosome of the individual
  for(ii = 0; ii < chrom_length; ii++)
    if (Flip(mutat_prob))
      {
        individual->mutate(ii, gene_low[ii], gene_high[ii]);
        mutations++;
      }
  
  // Check and correct any "illegal" mutations
  do 
    {
      flag = check_chrom(individual);
      if (flag) fix_chrom(individual, flag);
    } while (flag);
}

void POPULATION::statistics(const char *_stat_file)
{
  // Print to file the information about the current generation
  
  FILE *stat_file;
  int ii, jj, kk; // Index variables for the loops

  stat_file = fopen(_stat_file, "a");
  fprintf(stat_file,"\nGeneration %d", generation);
  fprintf(stat_file,"\nMember  Chromosome          Traits      ");
  fprintf(stat_file,"    Fitness   Parents"); 
  
  for(ii = 0; ii < pop_size; ii++)
    {
      fprintf(stat_file,"\n%3d    ", ii);
      for (jj = 0; jj < chrom_length; jj++)
	{  
	  for (kk = 1; kk < num_traits; kk++)
	    {
	      if (jj == trait_start[kk])
		{
		  fprintf(stat_file," ");
		}
	      fprintf(stat_file,"%d", members[ii]->gene[jj]);
	    }
	}
      fprintf(stat_file," ");
      for (jj = 0; jj < num_traits; jj++)
	{
	  fprintf(stat_file," %8.5f", members[ii]->trait_val[jj]);
	}
      fprintf(stat_file,"  %9.5f ", members[ii]->fitness);
      fprintf(stat_file,"%3d %3d", 
	      members[ii]->parents[0],
	      members[ii]->parents[1]);
    }
  
  fprintf(stat_file,"\nBest: %5.4f (#%d)",
	  members[bestmember]->fitness,  bestmember);
  
  fprintf(stat_file,"\nAverage: %5.4f",
	  sumfitness/(double)pop_size);
  fprintf(stat_file,"\nWorst: %5.4f (#%d)",
	  members[worstmember]->fitness, worstmember);
  fprintf(stat_file,"\nCrossovers: %d ", crossovers);
  fprintf(stat_file,"\nMutations: %d", mutations);
  fprintf(stat_file,"\n");
  fclose(stat_file);
}


void POPULATION::best_statistics(const char *_best_stat_file)
{
  // Print to file the information about the best member in the 
  // current generation
  
  FILE *best_stat_file;
  int jj, kk; // Index variables for the loops
  

  best_stat_file = fopen(_best_stat_file, "a");

  if (generation == 1)
    {
      fprintf(best_stat_file,
	      "\nGeneration :  Fitness   :    Chromosome    :   Traits\n");
    }
  fprintf(best_stat_file," %3d  ", generation);
  fprintf(best_stat_file,"%15.10f     ", members[bestmember]->fitness);
  for (jj = 0; jj < chrom_length; jj++)
    {  
      for (kk = 1; kk < num_traits; kk++)
	{
	  if (jj == trait_start[kk])
	    {
	      fprintf(best_stat_file," ");
	    }
	  fprintf(best_stat_file,"%d", members[bestmember]->gene[jj]);
	}
    }
  fprintf(best_stat_file,"  ");
  for (jj = 0; jj < num_traits; jj++)
    {
      fprintf(best_stat_file," %8.5f", members[bestmember]->trait_val[jj]);
    }
  
  fprintf(best_stat_file,"\n");
  fclose(best_stat_file);
}

void POPULATION::save_generation(const char *_generation_file)
{
  // Save all the data in the population structure to a file.
  
  FILE *generation_file;
  int ii, jj; // Index variables for the loops
  
  generation_file = fopen(_generation_file, "w");
  
  fprintf(generation_file,"Total Fitness: %20.15f \n", sumfitness);
  fprintf(generation_file,"Average Fitness: %20.15f \n", 
	  sumfitness/(double)pop_size);
  fprintf(generation_file,"Num of the best member: %d \n", bestmember);
  fprintf(generation_file,"Num of the worst member: %d \n", worstmember);
  fprintf(generation_file,"Population size: %d \n", pop_size);
  fprintf(generation_file,"Generation number: %d \n", generation);
  fprintf(generation_file,"Number of crossovers: %d \n", crossovers);
  fprintf(generation_file,"Number of mutations: %d \n\n", mutations);
  
  for(ii = 0; ii < pop_size; ii++)
    {
      fprintf(generation_file,"Member: %d \n", ii);
      fprintf(generation_file,"Chromosome: ");
      for (jj = 0; jj < chrom_length; jj++)
	{  
	  fprintf(generation_file,"%d ", members[ii]->gene[jj]);
	}
      fprintf(generation_file,"\n");
      fprintf(generation_file,"Traits: ");
      for (jj = 0; jj < num_traits; jj++)
	{  
	  fprintf(generation_file,"%15.10f ", members[ii]->trait_val[jj]);
	}
      fprintf(generation_file,"\n");
      fprintf(generation_file,"Fitness: %15.10f \n", members[ii]->fitness);
      fprintf(generation_file,"Parents: %d %d\n\n", 
	      members[ii]->parents[0],
              members[ii]->parents[1]);
    }
  
  fclose(generation_file);
}
